
;;; thingatpt-utils-base.el --- thing-at-point edit functions

;; Version: 1.5

;; Copyright (C) 2006, 2007 Andreas Roehler

;; Author: Andreas Roehler <andreas.roehler@easy-emacs.de>

;; This file is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; This file is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to
;; the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
;; Boston, MA 02110-1301, USA.

;; Keywords: convenience

;;; Changes to previous version: 

;; New function `delimited-atpt' handles regexps as
;; /[Hh][Rr][Ee][Ff]/

;;; Commentary: 

;; A set of functions to return, mover over or
;; manipulate a given THING. THING may be a well known
;; form as symbol', `list', `sexp', `defun' but also a
;; new defined and abstract thing. 

;; The idea comes from Mike Williams
;; <mikew@gopher.dosli.govt.nz>, author of
;; thingatpt.el

;; The goal is to have a set of similar forms. For
;; example, to provide a word with double-quotes around
;; it, call doublequote-word-atpt. In a similar way you
;; may double-quote not just a word, but any object
;; instrumented here as THING. To make parentheses
;; around it call parenthese-word-atpt, etc.

;; Move-functions of this package differ from common
;; behaviour in such, as `forward-word' i.e. stops
;; _not_ _after_ THING, but on the last char of
;; THING. That's in order to enable a call of
;; thing-at-point functions at the end
;; position. Otherwise, when cursor stops after word
;; (THING) finally, word-atpt would return nil. So it
;; returns word, boundaries of word or whatever is
;; called (and implemented).

;; To see other features, maybe try separate-list-atpt
;; or comment-list-atpt while point is inside a
;; list. Try it again with an abstract char-class as
;; [:alnum:], i.e. try comment-alnum-atpt,
;; brace-alnum-atpt etc.

;; Call `list-of-things-atpt' to see which objects are
;; presently instrumented, beside `delimited-atpt', where
;; selection specified on the fly (because of the wide range of
;; chars, designing such a possible THING).

;; Most THINGS are provided with a set of functions at,
;; before and after point - i.e. to call with
;; ACTION-THING-atpt, -bfpt, afpt.

;; This utility comes with test-functions which return
;; the possible results of most functions (exception
;; are the kill-fns). Call thatpt-test, thatpt-mv-test
;; or thatpt-delimtest over text. Thatpt-delimtest
;; changes but restores the buffer. Customize the speed
;; of execution via `thatpt-delimtest-delay' and
;; `thatpt-mv-test-delay.'

;; Diffs to basics of required thingatpt.el:
;; `bounds-of-thing-at-point' is replaced by a new
;; `bounds-of-single-thatpt', which now first searches
;; backward. As a consequence several
;; `beginning-op-atpt' and `end-op-atpt' constructs had
;; to be rewritten.

;; Behavior in general is not validating; i.e. if you
;; call url-atpt and there is no url, all chars at
;; point may be picked, which could be part of a
;; url. Sometimes, however, a kind of validation may be
;; introduced.

;; In case of trouble, please send me a bug report. Any
;; ideas and comments welcome.




;; How it works:

;; Thing-at-point delivers a portion of the
;; buffer. This substring is determined by two
;; alternative ways:

;; - If a pair of move-functions is known, as forward-
;;   and backward-word, its used. - A move-function
;;   specified for thingatpt, called beginning-op-atpt
;;   and end-op-atpt, may exist. The latter case given,
;;   this form will be used preferential. The point is
;;   stored after move. Beginning and end are delivered
;;   as pair: as consed bounds-of-thing. It's easy to
;;   write your own thing-at-point functions. You need
;;   three forms:
;; 
;; (defun MY-FORM-atpt (&optional arg ispec) 
;;   " "
;;   (interactive "P\np")
;;   (thatpt 'MY-FORM arg ispec))
;; 
;; (put 'MY-FORM 'beginning-op-atpt 
;;            (lambda () MY-FORWARD-MOVE-FUNKTION))
;; (put 'MY-FORM 'end-op-atpt
;;      (lambda () MY-BACKWARD-MOVE-FUNKTION))

;; For example if you want to pick all chars at point
;; which are written between a string "AAA" and a
;; "BBB", which may exist as
;; AAA Luckily detected a lot of things! BBB
;; After evaluation of
;; (put 'MY-FORM 'beginning-op-atpt
;;      (lambda ()
;;        (search-backward "AAA" nil t 1)
;;        ;; step chars of search expression back
;;        (forward-char 3)))
;; 
;; (put 'MY-FORM 'end-op-atpt
;;      (lambda ()
;;        (search-forward "BBB" nil t 1)
;;        (forward-char -3)))
;; together with the functions definition above, it's ready.
;; M-x MY-FORM-atpt 
;; (while point inside) you should see: 
;; " Luckily detected a lot of things! " 
;; in the minibuffer.

;; Todo: Enable operation over a given number of things forward
;; or backward from point. (The form, which will take the numeric
;; argument is given already, but has no effect at the moment.)

;;; Code:

(require 'thingatpt)

(defvar thatpt-orig 0
  "Correct orig according to delimiter-length")


;; Ascii

(put 'ascii 'beginning-op-atpt
     (lambda ()
       (when
         (looking-at "[[:ascii:]]")
       (skip-chars-backward "[:ascii:]"))))

(put 'ascii 'end-op-atpt
     (lambda ()
       (skip-chars-forward "[:ascii:]")))


;; Alnum

(put 'alnum 'beginning-op-atpt
     (lambda ()
       (when
         (looking-at "[[:alnum:]]")
       (skip-chars-backward "[:alnum:]"))))

(put 'alnum 'end-op-atpt
     (lambda ()
       (skip-chars-forward "[:alnum:]")))


;; Alpha

(put 'alpha 'beginning-op-atpt
     (lambda ()
       (when
         (looking-at "[[:alpha:]]")
       (skip-chars-backward "[:alpha:]"))))

(put 'alpha 'end-op-atpt
     (lambda ()
       (skip-chars-forward "[:alpha:]")))


;; Blank

(put 'blank 'beginning-op-atpt
     (lambda ()
       (when
         (looking-at "[[:blank:]]")
       (skip-chars-backward "[:blank:]"))))

(put 'blank 'end-op-atpt
     (lambda ()
       (skip-chars-forward "[:blank:]")))


;; Cntrl

(put 'cntrl 'beginning-op-atpt
     (lambda ()
       (when
         (looking-at "[[:cntrl:]]")
       (skip-chars-backward "[:cntrl:]"))))

(put 'cntrl 'end-op-atpt
     (lambda ()
       (skip-chars-forward "[:cntrl:]")))


;; Digit

(put 'digit 'beginning-op-atpt
     (lambda ()
       (when
         (looking-at "[[:digit:]]")
       (skip-chars-backward "[:digit:]"))))

(put 'digit 'end-op-atpt
     (lambda ()
       (skip-chars-forward "[:digit:]")))


;; Graph

(put 'graph 'beginning-op-atpt
     (lambda ()
       (when
         (looking-at "[[:graph:]]")
       (skip-chars-backward "[:graph:]"))))

(put 'graph 'end-op-atpt
     (lambda ()
       (skip-chars-forward "[:graph:]")))


;; Lower

(put 'lower 'beginning-op-atpt
     (lambda ()
       (when
         (looking-at "[[:lower:]]")
       (skip-chars-backward "[:lower:]"))))

(put 'lower 'end-op-atpt
     (lambda ()
       (skip-chars-forward "[:lower:]")))


;; Multibyte

(put 'multibyte 'beginning-op-atpt
     (lambda ()
       (when
         (looking-at "[[:multibyte:]]")
       (skip-chars-backward "[:multibyte:]"))))

(put 'multibyte 'end-op-atpt
     (lambda ()
       (skip-chars-forward "[:multibyte:]")))


;; Nonascii

(put 'nonascii 'beginning-op-atpt
     (lambda ()
       (when
         (looking-at "[[:nonascii:]]")
       (skip-chars-backward "[:nonascii:]"))))

(put 'nonascii 'end-op-atpt
     (lambda ()
       (skip-chars-forward "[:nonascii:]")))


;; Print

(put 'print 'beginning-op-atpt
     (lambda ()
       (when
         (looking-at "[[:print:]]")
       (skip-chars-backward "[:print:]"))))

(put 'print 'end-op-atpt
     (lambda ()
       (skip-chars-forward "[:print:]")))


;; Punct

(put 'punct 'beginning-op-atpt
     (lambda ()
       (when
         (looking-at "[[:punct:]]")
       (skip-chars-backward "[:punct:]"))))

(put 'punct 'end-op-atpt
     (lambda ()
       (skip-chars-forward "[:punct:]")))


;; Space

(put 'space 'beginning-op-atpt
     (lambda ()
       (when
         (looking-at "[[:space:]]")
       (skip-chars-backward "[:space:]"))))

(put 'space 'end-op-atpt
     (lambda ()
       (skip-chars-forward "[:space:]")))


;; Unibyte

(put 'unibyte 'beginning-op-atpt
     (lambda ()
       (when
         (looking-at "[[:unibyte:]]")
       (skip-chars-backward "[:unibyte:]"))))

(put 'unibyte 'end-op-atpt
     (lambda ()
       (skip-chars-forward "[:unibyte:]")))


;; Upper

(put 'upper 'beginning-op-atpt
     (lambda ()
       (when
         (looking-at "[[:upper:]]")
       (skip-chars-backward "[:upper:]"))))

(put 'upper 'end-op-atpt
     (lambda ()
       (skip-chars-forward "[:upper:]")))


;; Word

(put 'word 'beginning-op-atpt
     (lambda ()
       (when
         (looking-at "[[:word:]]")
       (skip-chars-backward "[:word:]"))))

(put 'word 'end-op-atpt
     (lambda ()
       (skip-chars-forward "[:word:]")))


;; Xdigit

(put 'xdigit 'beginning-op-atpt
     (lambda ()
       (when
         (looking-at "[[:xdigit:]]")
       (skip-chars-backward "[:xdigit:]"))))

(put 'xdigit 'end-op-atpt
     (lambda ()
       (skip-chars-forward "[:xdigit:]")))



(setq thatpt-forms-list '(
			  alnum
			  alpha
			  ascii
			  blank
			  cntrl
			  csv
			  date
			  defun
			  delimited
			  digit
			  email
			  filename
			  float
			  graph
			  isbn
			  line
			  list
			  lower
			  ml-text
			  multibyte
			  nonascii
			  number
			  page
			  paragraph
			  phone
			  print
			  punct
			  region
			  sentence
			  sexp
			  space
			  string
			  symbol
			  unibyte
			  upper
			  url
			  whitespace
			  word
			  xdigit
			  ))

(defconst thatpt-version "1.5" "\
Version numbers of this version of thingatpt-utils.")

(defun thatpt-version (&optional ispec) 
  "Print and/or return thatpt-version"
  (interactive "p")
  (when ispec
    (message "%s" thatpt-version))
  thatpt-version)

;;; CSV

;; Value of var `csv-separators' will be taken according to
;;; csv-mode.el --- major mode for editing comma-separated value files
;; Author: Francis J. Wright <F.J.Wright at qmul.ac.uk>
;; URL: http://centaur.maths.qmul.ac.uk/Emacs/

(defcustom atpt-separator ";"
  "Char to distinguish datasets in a `comma`-separated row" 
  :type 'string
  :group 'convenience)

(when (boundp 'csv-separators)
    (setq separator-atpt csv-separators))

(put 'csv 'beginning-op-atpt
     (lambda ()
       (skip-chars-backward (concat "^" (car separator-atpt)) (line-beginning-position))))

(put 'csv 'end-op-atpt
     (lambda ()
       (skip-chars-forward (concat "^" (car separator-atpt))(line-end-position))))

;;; ISBN

(put 'isbn 'beginning-op-atpt
     (lambda ()
       (unless (looking-at "\\s-")
	 (skip-chars-backward "[0-9]-"))))

(put 'isbn 'end-op-atpt
     (lambda ()
       (when (looking-at "[0-9]\\{1,3\\}[0-9-]\\{7,12\\}[0-9X]\\{0,1\\}")
	 (goto-char (match-end 0)))))

;;; Symbol

(put 'symbol 'beginning-op-atpt
     (lambda ()
       (unless (looking-at "\\s-")
	 (skip-syntax-backward "W_"))))

(put 'symbol 'end-op-atpt
     (lambda ()
       (skip-syntax-forward "W_")))

;; DATE

(put 'date 'beginning-op-atpt
     (lambda ()
       ;; provide for the case, we are over a
       ;; string-delimiter as `"'
       (when
	   (and (not (eq 32 (char-after))) 
		(or (bobp)
		    (eq 32 (char-before))))
	 (forward-char 1)
	 ;; as the bounds-function checks position, correct it
	 (setq thatpt-orig 1)) 
       (skip-chars-backward "[0-9] .-")
       (skip-chars-forward " ") 
))

(put 'date 'end-op-atpt
     (lambda ()
              (skip-chars-forward "[0-9] .-")
	      (skip-chars-backward " ")))

;; Url

(put 'url 'beginning-op-atpt
     (lambda ()
       ;; provide for the case, we are over a
       ;; string-delimiter as `"'
       (when
	   (and (not (eq 32 (char-after))) 
		(or (bobp)
		    (eq 32 (char-before))))
	 (forward-char 1)
	 ;; as the bounds-function checks position, correct it
	 (setq thatpt-orig 1)) 
       (skip-chars-backward ":/?#[]@!$&()*+,;=[:alnum:]-._~")
))

(put 'url 'end-op-atpt
     (lambda ()
       	 (skip-chars-forward ":/?#[]@!$&()*+,;=[:alnum:]-._~")
	 (skip-chars-backward ":")))

;; Phone

(put 'phone 'beginning-op-atpt
     (lambda ()
       (when
	   (and (looking-at "[0-9 \t.()-]")
		(not (eq (char-before) ?+)))
       (re-search-backward "[^0-9 \t.()-][0-9 ()\t-]+" (line-beginning-position) t 1) (forward-char 1)))) 

(put 'phone 'end-op-atpt
     (lambda ()
       (when
	   (looking-at "[0-9;, \t()-]")
	 (re-search-forward "[0-9 \t.()-]+[^0-9 \t-]" (1+ (line-end-position)) t 1) (forward-char -1)))) 

;; Text
;; Useful to extract texts between ml-tags

(put 'ml-text 'beginning-op-atpt
     (lambda ()
       (when
	   (looking-at "[^>]") 
	 (re-search-backward ">" nil t 1)
         (forward-char 1)))) 

(put 'ml-text 'end-op-atpt (lambda () (re-search-forward "</" nil t 1) (forward-char -2))) 

;; Email

(put 'email 'beginning-op-atpt
     (lambda ()
       (when
	   (looking-at "[^ \t]")
	 (re-search-backward "[,;][[:graph:]]\\|<[[:graph:]]\\|^[[:graph:]]\\|[^[:graph:]][[:graph:]]" (line-beginning-position) t 1)(when (looking-at "[[:space:];,]") (forward-char 1))))) 

;; (put 'email 'end-op-atpt (lambda () (re-search-forward "[[:graph:]]+>\\|[[:graph:]]+@[[:graph:]]+[> \t\n]*" (line-end-position) t 1)))

(put 'email 'end-op-atpt (lambda () (when (looking-at "[ <]\\{0,1\\}\\([\041-\132\136-\176]+@[\041-\132\136-\176]+\\)[;,> \t\n]*")
				 (goto-char (match-end 1))
				 (skip-chars-backward "[[:punct:]]")))) 

			   ;; (put 'email 'end-op-atpt (lambda () (when (looking-at "[ <]\\{0,1\\}\\([[:graph:]]+@[[:graph:]]+\\)[;,> \t\n]*")
			   ;; 			   (goto-char (match-end 1))
			   ;; 			   (skip-chars-backward "[[:punct:]]"))))
			   
;; Whitespace

(put 'whitespace 'beginning-op-atpt (lambda () (when (looking-at "[ \t]") (skip-chars-backward "[:blank:]"))))

(put 'whitespace 'end-op-atpt (lambda () (skip-chars-forward "[:blank:]")))

;; Number

(put 'number 'beginning-op-atpt
     (lambda ()
       (when
	   (numberp (string-to-number (buffer-substring-no-properties (point) (1+ (point)))))
	 (skip-chars-backward "[0-9]"))))

;; (put 'number 'beginning-op-atpt (lambda () (when (numberp (read (buffer-substring-no-properties (point) (1+ (point)))))
;;     (skip-chars-backward "[0-9]"))))

(put 'number 'end-op-atpt
 (lambda ()
   (skip-chars-forward "[0-9]")))

;; Floats

(put 'float 'beginning-op-atpt (lambda () (when (numberp (read (buffer-substring-no-properties (point) (1+ (point)))))
    (skip-chars-backward "[0-9].,"))))

(put 'float 'end-op-atpt (lambda () (skip-chars-forward "[0-9.,]")))

;; Sexp

;; (put 'sexp 'beginning-op-atpt
;;      (lambda ()
;;        (unless (or (looking-at "\\s-")
;; 		   (looking-back "\\s-"))
;; 	 (backward-sexp))))
;; 
;; (put 'sexp 'end-op-atpt
;;      (lambda ()
;;        (forward-sexp)))

;; Filename

(put 'filename 'beginning-op-atpt
     (lambda ()
       (re-search-backward (concat "[^" thing-at-point-file-name-chars "]") nil t)
       (forward-char 1)))

(put 'filename 'end-op-atpt
     (lambda ()
       (re-search-forward (concat "\\=[" thing-at-point-file-name-chars "]*")
			  nil t)
     (skip-chars-backward ": ")))

;; Defun

(put 'defun 'beginning-op-atpt (lambda (&optional arg) (beginning-of-defun (or arg 1))))

(put 'defun 'end-op-atpt (lambda (&optional arg)(end-of-defun (or arg 1))))

;; Lines

(put 'line 'beginning-op-atpt (lambda () (beginning-of-line)))

;; Strings

(put 'string 'beginning-op-atpt
     (lambda ()
       (let ((pos (with-syntax-table (standard-syntax-table) (nth 8 (syntax-ppss)))))
	 (when pos
	   (goto-char pos)))
;; (forward-char 1)
))

(put 'string 'end-op-atpt
     (lambda ()
       (let ((pos (progn (save-excursion (beginning-of-defun) (point))))) 
	 (forward-char 1) 
	 (while (not (eq (char-after) (nth 3 (with-syntax-table (standard-syntax-table) (parse-partial-sexp pos (point))))))
	   (forward-char 1)))
       (forward-char 1))) 
   
;;  Lists  

(put 'list 'end-op-atpt (lambda () (forward-list 1)
)) 

(put 'list 'beginning-op-atpt
     (lambda ()
       (unless (looking-at "\\s(")
	 (if (nth 9 (syntax-ppss))
	     (goto-char (car (last (nth 9 (syntax-ppss)))))
	   (backward-sexp)
	   (forward-char -1)))))

;; (put 'list 'beginning-op-atpt
;;      (lambda ()
;;        (or (looking-at "\\s(")
;; 	   (when (nth 9 (syntax-ppss))
;; 	     (goto-char (car (last (nth 9 (syntax-ppss)))))))))

;; Region

(put 'region 'end-op-atpt
     (lambda () 
       (goto-char (region-end))))

(put 'region 'beginning-op-atpt
     (lambda ()
       (goto-char (region-beginning))))


;; Delimited

(put 'delimited 'beginning-op-atpt
     (lambda ()
       (when
	   (looking-at (concat "[^" delimiter-atpt "]"))
	 (skip-chars-backward (concat "^" delimiter-atpt)))
       (when with-delimiters-atpt
	 (forward-char -1))))

(put 'delimited 'end-op-atpt
     (lambda ()
       (when with-delimiters-atpt
	 (forward-char 1))
       (skip-chars-forward (concat "^" delimiter-atpt))
       (when with-delimiters-atpt
	 (forward-char 1))))

(defcustom with-delimiters-atpt  t 
 "If containing delimiters are copied with string" 

:type 'boolean
:group 'convenience
)

(defcustom delimiter-atpt "/"
  "Specify the delimiter char."
  :type 'string
  :group 'convenience)



(defun list-of-things-atpt ()
  "Displays a list of objects which might be called as THING herewith
  Every THING is provided with a set of functions at, before and after point 
  - i.e. to call with ACTION-THING-atpt,  -bfpt, afpt."
  (interactive)
  (message "%s" thatpt-forms-list))

(defun thatpt (thing &optional arg ispec) 
  "Returns a buffer substring according to THING.
  THING may be a well known form as `symbol',
  `list', `sexp', `defun'.
  You may also define new and abstract kinds of THING.
  See example given in thingatpt-utils-base.el.
  Further functions with `thatpt' provide moves, transpositions.  
  Call `list-of-things-atpt' to see what's implemented.
  "
  (or arg (setq arg 1))
  ;; Special handling of `delimited-atpt':
  ;; As a wide range of delimiters are feasible,
  ;; specify them on the fly.
  (when (and (eq thing 'delimited) (or (stringp arg) (< 1 arg)))
    (if (stringp arg)
	(setq delimiter-atpt arg)
      (when ispec
	(setq delimiter-atpt
	      (read-from-minibuffer "Delimiter-atpt: " delimiter-atpt))
	;;       (setq delimiter-atpt delimiter))
	))
    (put 'delimited 'beginning-op-atpt
	 (lambda ()
	   (when
	       (looking-at (concat "[^" delimiter-atpt "]"))
	     (skip-chars-backward (concat "^" delimiter-atpt)))
	   (when with-delimiters-atpt
	     (backward-char 1))))
    (put 'delimited 'end-op-atpt
	 (lambda ()
	   (when with-delimiters-atpt
	     (forward-char 1))
	   (skip-chars-forward (concat "^" delimiter-atpt))
	   (when with-delimiters-atpt
	     (forward-char 1)))))
  (condition-case nil
      (let* ((bounds-at-point (thatpt-bounds thing arg ispec))
	     (type (buffer-substring-no-properties (car bounds-at-point) (cdr bounds-at-point))))
	(when ispec
	  (kill-new type)
	  (message "   %s" type))
	type)
    ;;     (error nil)
    (error nil)))

(defun bounds-of-single-thatpt (thing &optional arg backward-flag)
  "Determine the start and end buffer locations for the THING at point.
  THING is a symbol which specifies the kind of syntactic entity you want.
  Possibilities include `symbol', `list', `sexp', `defun', `filename', `url',
  `word', `sentence', `whitespace', `line', `page' and others.
  Call THING by his name, i.e. word-atpt etc., see `list-of-things-atpt' to see what's implemented"
  (condition-case nil
      (save-excursion
	(let ((orig (point))
	      (thatpt-orig 0)
	      (beg (progn 
		     (funcall ;; First, move to beg.
		      (or (get thing 'beginning-op-atpt) 
			  (lambda ()
			    (forward-char 1)
			    (forward-thing thing -1))))
		     (point)))
	      (end
	       (progn (funcall ;; Then move to end.
		       (or (get thing 'end-op-atpt)
			   (lambda () (forward-thing thing 1))))
		      (point))))
	  (when (< beg end)
	    (cons beg end))))
    (error nil)))

(defun thatpt-bounds (thing &optional arg ispec)
  (or arg (setq arg 1))
  (condition-case nil 
      (save-excursion 
	(let* ((move-fn (if (< 0 arg)
			    (intern-soft (concat "forward-"(symbol-name thing)"-atpt"))
			  (intern-soft (concat "backward-"(symbol-name thing)"-atpt"))))
	       (orig-bounds (bounds-of-single-thatpt thing))
	       (abs-arg (abs arg))
	       (bounds (if (< 1 abs-arg)
			   (progn
			     (funcall move-fn abs-arg)
			     (bounds-of-single-thatpt thing))
			 orig-bounds))
	       (extended-bounds (cond ((and bounds (< 0 arg) (string-match "forward" (format "%s" move-fn)))
				       (cons (car orig-bounds) (cdr bounds)))
				      ((and bounds (> 0 arg) (string-match "backward" (format "%s" move-fn)))
				       (cons (cdr orig-bounds) (car bounds)))
				      (t	nil))))
	  (if extended-bounds
	      (progn
		(when ispec (message "%s" extended-bounds))
		extended-bounds)
	    orig-bounds)))
    (error nil)))

(defun thatpt-beginning (thing &optional arg ispec)
  (condition-case nil 
      (let* ((bounds (bounds-of-single-thatpt thing arg))
	     (start (car bounds)))
	(when ispec
	  (message "   %s " start)) 
	start)
    (error nil)))

(defun thatpt-end (thing &optional arg ispec)
  (condition-case nil 
      (let* ((bounds (bounds-of-single-thatpt thing arg))
	     (end (cdr bounds)))
	(when ispec
	  (message "   %s "  end)) 
	end)
    (error nil)))

(defun thatpt-escape (thing)
  " "
  (regexp-quote (buffer-substring-no-properties (thatpt-beginning thing)(thatpt-end thing))))

(defun thatpt-check (beg end &optional arg ispec)
  "Count letters, number etc. at point. Relate them."
  (let*
      (
       ;;     (bounds (bounds-of-single-thatpt thing arg))
       ;;     (beg (car bounds))
       ;;     (end (cdr bounds))
       (refchars
	(if
	    (string-match "[/-]" (buffer-substring-no-properties beg end))
	    t
	  nil))
       (dots
	(if
	    (string-match "[\.]" (buffer-substring-no-properties beg end))
	    t
	  nil))
       (letters (count-matches "[a-z-A-ZäöüÄÖÜß]" beg end))
       (numbers (count-matches "[0-9]" beg end))
       (printables (count-matches "[[:print:]]" beg end))
       (alph-num-ratio
	(when (< 0 numbers)
	  (/ letters numbers)))
       (alph-num-rest
	(when (< 0 numbers)
	  (% letters numbers))))
    (when ispec
      (message "letters numbers printables alph-num-ratio alph-num-rest refchars dots beg end \n %s %s %s %s %s %s %s %s %s" letters numbers printables alph-num-ratio alph-num-rest refchars dots beg end))
    (list letters numbers printables alph-num-ratio alph-num-rest refchars dots beg end)))

(defun thatpt-copy (thing &optional arg ispec)
  (condition-case nil 
      (let ((newcopy (thatpt thing arg)))
	(if newcopy
	    (progn
	      (kill-new newcopy)
	      (if ispec
		  (message "   %s" (car kill-ring))
		(car kill-ring)))
	  nil))
    (error nil)))

(defun thatpt-separate (thing &optional arg ispec) 
  " "
  (condition-case nil 
      (let* ((bounds (thatpt-bounds thing arg))
	     (beg (car bounds))
	     (end (if (listp (cdr bounds))
		      (cadr bounds)
		    (cdr bounds)))
	     (oldbufsize (buffer-size)))
	(goto-char beg)
	(when (and beg end (<= (line-beginning-position) beg) (not (looking-back "^[ \t]+")))
	  (goto-char beg)
	  (newline-and-indent))
	(save-excursion 
	  (when
	      (< oldbufsize (buffer-size)) 
	    (setq end (+ end (- (buffer-size) oldbufsize)))
	    (setq beg (+ beg (- (buffer-size) oldbufsize)))
	    (setq oldbufsize (buffer-size)))
	  (goto-char end)
	  (cond ((eobp)
		 (newline-and-indent))
		((looking-at "[\t\r\n\f ]*$")
		 nil)
		(t (newline-and-indent)))
	  (list beg end)))
    (error nil)))

(defun thatpt-comment (thing &optional arg ispec) 
  " "
  (condition-case nil 
      (let* ((bounds (thatpt-separate thing arg ispec))
	     (beg (car bounds))
	     (end (cadr bounds)))
	(if (and beg end)
	    (progn 
	      (goto-char beg)
	      (comment-or-uncomment-region beg (1+ end)))
	  nil))
    (error nil))) 

(defun thatpt-kill (thing &optional arg) 
  " "
  (condition-case nil 
      (let* ((arg (or arg 1))
	     (bounds (bounds-of-single-thatpt thing arg)) 
	     (start (car bounds))
	     (end (cdr bounds)))
	(kill-region start end))
    (error nil)))

(defun thatpt-mark (thing &optional arg) 
  " "
  (condition-case nil 
      (let* ((arg (or arg 1))
	     (bounds (bounds-of-single-thatpt thing arg))) 
	(transient-mark-mode t)
	(goto-char (car bounds))
	(push-mark)
	(goto-char (cdr bounds))
	(exchange-point-and-mark))
    (error nil)))

(defun thatpt-forward (thing &optional arg ispec) 
  " "
  (or arg (setq arg 1))
  (condition-case nil
      (if (< 0 arg)
	  (while (< 0 arg)
	    ;; otherwise containing list is taken
	    (when (and (eq thing 'list) (not (eq 4 (syntax-class (syntax-after (point)))))) 
	      (skip-syntax-forward "^("))
	    (let ((orig (point)) 
		  (ep (cdr (bounds-of-single-thatpt thing))))
	      ;; if inside THING, go to the end
	      (if (and ep (< orig (1- ep)))
		  (goto-char (1- ep))
		;; if not over or at the end of THING, go to next
		(unless (eobp)
		  (forward-char 1))
		;; otherwise containing list is taken
		(when (and (eq thing 'list) (not (eq 4 (syntax-class (syntax-after (point))))))
		  (skip-syntax-forward "^("))
		(when (eq ep (point))
		  (forward-char 1)) 
		(while (and (not (bobp))(not (bounds-of-single-thatpt thing)))
		  (forward-char 1)))
	      ;; go to the last char of THING
	      (goto-char (1- (cdr (bounds-of-single-thatpt thing))))
	      (setq arg (1- arg))))
	;; (> 0 arg)
	(while
	    (> 0 arg)
	  ;; otherwise containing list is taken
	  (when (and (eq thing 'list) (not (eq 5 (syntax-class (syntax-after (point)))))) 
	    (skip-syntax-backward "^)")
	    (forward-char -1))
	  (let ((orig (point)) 
		(ap (car (bounds-of-single-thatpt thing))))
	    (if (and ap (> orig ap))
		(goto-char ap)
	      ;; we are not over or at the end of THING
	      (unless (bobp)
		(forward-char -1))
	      (while (and (not (bobp)) (not (bounds-of-single-thatpt thing)))
		(forward-char -1)))
	    (goto-char (car (bounds-of-single-thatpt thing)))
	    (setq arg (1+ arg)))))
    (when ispec (message "%s" (point))) 
    (point)
    (error nil)))

(defun thatpt-backward (thing &optional arg ispec) 
  " "
  (or arg (setq arg 1))
  (condition-case nil
      (thatpt-forward thing (- (or arg 1)))
    (when ispec (message "%s" (point)))
    (point)
    (error nil)))

(defun thatpt-transpose (thing &optional arg)
  " "
  (let ((pos1 (or (funcall (car (read-from-string (concat "bounds-of-" (symbol-name thing) "-atpt"))) arg)
		  (funcall (car (read-from-string (concat "bounds-of-" (symbol-name thing) "-afpt"))) arg)))
	(pos2 (funcall (car (read-from-string (concat "bounds-of-" (symbol-name thing) "-bfpt"))) arg)))
    (message "Transpose objects \w boundaries: %s %s" pos1 pos2)
    (transpose-subr-1 pos1 pos2)))

;; transpose-subr-1
;; (8 . 11) (2 . 7)
;; Mit Arg.: Puffer merken
;; thatpt-transpose - Fehler
;; (6 . 11) (2 . 5)

;; slightly changed from simple.el

(defun transpose-subr-ar (mover arg &optional special)
  (let ((aux (if special mover
	       (lambda (x)
		 (cons (progn (funcall mover x) (point))
		       (progn (funcall mover (- x)) (point))))))
	pos1 pos2)
    (cond
     ((= arg 0)
      (save-excursion
	(setq pos1 (funcall aux 1))
	(goto-char (mark))
	(setq pos2 (funcall aux 1))
	(transpose-subr-1 pos1 pos2))
      (exchange-point-and-mark))
     ((> arg 0)
      (setq pos1 (funcall aux -1))
      ;; 2007-04-12 a.roehler@web.de changed section start 
      ;; thingatpt stops over last char of thing, not behind
      (setcdr pos1 (1+ (cdr pos1)))
      ;; 2007-04-12 a.roehler@web.de changed section end 
      (setq pos2 (funcall aux arg))
      ;; 2007-04-12 a.roehler@web.de changed section start
      (setcar pos2 (1+ (car pos2)))
      ;; 2007-04-12 a.roehler@web.de changed section end 
      (transpose-subr-1 pos1 pos2)
      (goto-char (car pos2)))
     (t
      (setq pos1 (funcall aux -1))
      (goto-char (car pos1))
      (setq pos2 (funcall aux arg))
      (transpose-subr-1 pos1 pos2)))))

(defun thatpt-delim (thing action &optional arg ispec)
  "Process begin and end of region according to value of 
  `delim-action'
  If no region is active, process borders of THING-at-point
  according to value of delim-action-beginning- resp. -end-position
  Default is symbol-atpt.
  With \C-u or arg `escaped' to `t': insert escaped doublequotes"
  (or arg (setq arg 1))
  (condition-case nil 
      (save-excursion
	(let ((delim-insert (cond ((string= action "singlequote")
				   (cons ?\' ?\'))
				  ((string= action "doublequote")
				   (cons ?\" ?\"))
				  ((string= action "parentize")
				   (cons ?\( ?\)))
				  ((string= action "brace")
				   (cons ?\{ ?\}))
				  ((string= action "bracket")
				   (cons ?\[ ?\]))))
	      (oldbufsize (buffer-size))
	      (start (cond ((and mark-active transient-mark-mode)
			    (region-beginning))
			   ;; 		       (t (funcall (intern-soft (concat (symbol-name thing)"-atpt-beginning-position"))))))
			   (t (funcall (intern-soft (concat (symbol-name thing)"-atpt-beginning-position"))))))
	      (end (cond ((and mark-active transient-mark-mode)
			  (region-end))
			 (t (funcall (intern-soft (concat (symbol-name thing)"-atpt-end-position")))))))
	  (if start
	      (progn (goto-char start)
		     (insert (car delim-insert))
		     (if (< oldbufsize (buffer-size)) 
			 (setq end (+ end (- (buffer-size) oldbufsize)))
		       (setq end (- end (- oldbufsize (buffer-size)))))
		     (goto-char end)
		     (insert (cdr delim-insert))))))
    (error nil)))

(defun syntax-class-atpt (&optional ispec) 
  ""
  (interactive "p")
  (let ((sc (syntax-atpt)))
    (if ispec
	(message "%s" sc)
      sc)))

(defun syntax-class-bfpt (&optional ispec) 
  ""
  (interactive "p")
  (let ((sc (syntax-bfpt)))
    (if ispec
	(message "%s" sc)
      sc)))

(defun syntax-atpt (&optional ispec) 
  (interactive "p")
  (let ((stax (syntax-after (point))))
    (when ispec
      (message (format "%s" stax)))
    (car stax)))

(defun syntax-bfpt (&optional ispec) 
  (interactive "p")
  (let ((stax (syntax-after (1- (point)))))
    (when ispec
      (message (format "%s" stax)))
    (car stax)))

;; Until then when `up-list' was called while in
;;  string, it failed with message: 
;; Scan error: ... "Unbalanced parentheses" 

(defun up-list-atpt (&optional arg)
  "Move forward out of one level of parentheses.
With ARG, do this that many times.
A negative argument means move backward but still to a less deep spot."
  (interactive "p")
  (or arg (setq arg 1))
  (let ((inc (if (> arg 0) 1 -1)))
    (while (/= arg 0)
      (goto-char
       (or
	(progn
	  (when (in-string-p)
	    (goto-char (or (nth 8 (syntax-ppss)) (point)))
	    (forward-sexp 1))
	  (scan-lists (point) inc 1)) (buffer-end arg)))
      (setq arg (- arg inc)))))

(defun forward-list-at-point (&optional arg)
  "Move forward across one balanced group of parentheses.
With ARG, do it that many times.
Negative arg -N means move backward across N groups of parentheses."
  (interactive "p")
  (or arg (setq arg 1))
  (let ((pos (point))
	(uplist-arg (if (< 0 arg)
			1
		      -1))
	(akt-open (nth 1 (parse-partial-sexp 1 (point)))))
    (if (eq 4 (syntax-class (syntax-after (point))))
	(forward-list arg)
      (if (in-string-p)
	  (progn 
	    (goto-char (or (nth 8 (parse-partial-sexp 1 (point))) (point)))
	    (forward-char -1)
	    (up-list uplist-arg)
	    (if (< 0 arg)
		(setq arg (1- arg)) 
	      (setq arg (1+ arg))))
	(when (and (< 0 (abs arg)) (list-atpt))
	  (up-list uplist-arg)
	  (if (< 0 arg)
	      (setq arg (1- arg))
	    (setq arg (1+ arg)))))
      (forward-list arg))))

(defun delist-list-atpt ()
  "Delete delimiters of a list at point"
  (interactive "*")
  (let ((poss (bounds-of-list-atpt)))
    (narrow-to-region (car poss) (if (listp (cdr poss))
				     (cadr poss)
				   (cdr poss)))
    ;;     (narrow-to-region (car poss) (cadr poss))
    (point-min-marker)
    (point-max-marker) 
    (goto-char (point-min-marker))
    (delete-char 1)
    (goto-char (point-max-marker)) 
    (backward-delete-char 1)
    (widen)))


(provide 'thingatpt-utils-base)
;;; thingatpt-utils-base.el ends here
